
/* darkstat: a network traffic analyzer
 * (c) 2001-2002, Emil Mikulic.
 */

/* Network [ab]use graph code */

#include "graph.h"
#include <assert.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>

pthread_mutex_t graph_mutex;

i64 graph_sec_in[GRAPH_SECS], graph_sec_out[GRAPH_SECS];
i64 graph_min_in[GRAPH_MINS], graph_min_out[GRAPH_MINS];
i64 graph_hr_in[GRAPH_HRS], graph_hr_out[GRAPH_HRS];
i64 graph_day_in[GRAPH_DAYS], graph_day_out[GRAPH_DAYS];

int pos_sec = 0, pos_min = 0, pos_hr = 0, pos_day = 0;
int g_secs = 0, g_mins = 0, g_hrs = 0;
dword last_time = 0;



void init_graph(void)
{
	int i;
	i64 zero;
	SET64(zero, 0,0);

	for (i=0; i<GRAPH_SECS; i++)
		graph_sec_in[i] = graph_sec_out[i] = zero;
	for (i=0; i<GRAPH_MINS; i++)
		graph_min_in[i] = graph_min_out[i] = zero;
	for (i=0; i<GRAPH_HRS; i++)
		graph_hr_in[i] = graph_hr_out[i] = zero;
	for (i=0; i<GRAPH_DAYS; i++)
		graph_day_in[i] = graph_day_out[i] = zero;
}



inline void graph_add_in(const dword amount)
{
	add_64bit(graph_sec_in[pos_sec], amount);
	add_64bit(graph_min_in[pos_min], amount);
	add_64bit(graph_hr_in[pos_hr], amount);
	add_64bit(graph_day_in[pos_day], amount);
}

inline void graph_add_out(const dword amount)
{
	add_64bit(graph_sec_out[pos_sec], amount);
	add_64bit(graph_min_out[pos_min], amount);
	add_64bit(graph_hr_out[pos_hr], amount);
	add_64bit(graph_day_out[pos_day], amount);
}



inline void graph_pushone(void)
{
	i64 zero;
	SET64(zero, 0,0);

	pos_sec = (pos_sec + 1) % GRAPH_SECS;
	graph_sec_in[pos_sec] = graph_sec_out[pos_sec] = zero;

	g_secs++;
	if (g_secs == 60) /* this is a constant */
	{
		/* there are exactly 60 seconds in a minute, how many
		 * seconds you actually want to graph is your problem
		 */
		g_secs = 0;
		pos_min = (pos_min + 1) % GRAPH_MINS;
		graph_min_in[pos_min] = graph_min_out[pos_min] = zero;

		g_mins++;
		if (g_mins == 60) /* constant */
		{
			g_mins = 0;
			pos_hr = (pos_hr + 1) % GRAPH_HRS;
			graph_hr_in[pos_hr] = graph_hr_out[pos_hr] = zero;

			g_hrs++;
			if (g_hrs == 24) /* constant */
			{
				FILE *fp = fopen(daylog_file, "a");
				struct tm *tm;
				time_t tarm;

				tarm = time(NULL);
				tm = localtime(&tarm);

#ifdef HAVE_QUAD
			fprintf(fp, "%d/%d/%d|%Lu|%Lu\n",
					tm->tm_year+1900, tm->tm_mon+1,
					tm->tm_mday,
					graph_day_in[pos_day],
					graph_day_out[pos_day]);
#else
			{
				char *din, *dout;
				
				din = bignum_from_64bit(
						graph_day_in[pos_day]);
				dout = bignum_from_64bit(
						graph_day_out[pos_day]);

				fprintf(fp, "%d/%d/%d|%s|%s\n",
						tm->tm_year+1900, tm->tm_mon+1,
						tm->tm_mday, din, dout);

				free(din);
				free(dout);
			}
#endif
				fclose(fp);

				g_hrs = 0;
				pos_day = (pos_day + 1) % GRAPH_DAYS;
				graph_day_in[pos_day] =
					graph_day_out[pos_day] = zero;
			}
		}
	}
}



/* argument is ignored */
inline void graph_rotate(dword now)
{
	if (now == last_time) return;
	if (last_time == 0)
	{
		last_time = now;
		return;
	}

	for (; last_time<now; last_time++)
		graph_pushone();
}



inline void graph_save_one(FILE *fp, const int pos, const int GRAPH_SIZE,
		const int g, const i64 graph_in[], const i64 graph_out[])
{
	int i;

	fprintf(fp, "%u/%u %u\n", pos, GRAPH_SIZE, g);
	for (i=0; i<GRAPH_SIZE; i++)
		fprintf(fp,"%u:%u %u:%u\n",
				HI(graph_in[i]),  LO(graph_in[i]),
				HI(graph_out[i]), LO(graph_out[i]));
}



void graph_save(FILE *fp)
{
	graph_save_one(fp, pos_sec, GRAPH_SECS, g_secs,
			graph_sec_in, graph_sec_out);

	graph_save_one(fp, pos_min, GRAPH_MINS, g_mins,
			graph_min_in, graph_min_out);

	graph_save_one(fp, pos_hr , GRAPH_HRS , g_hrs ,
			graph_hr_in , graph_hr_out );

	graph_save_one(fp, pos_day, GRAPH_DAYS, 0     ,
			graph_day_in, graph_day_out);
}

