
/* Network traffic analyzer
 * (c) 2001-2002, Emil Mikulic.
 */

/* a somewhat functional mini-httpd */
#include "darkstat.h"
#include "www.h"
#include "dns.h"
#include "host_db.h"
#include "port_db.h"
#include "proto.h"
#include "graph.h"
#include "gif.h"
#include "content.h"

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>



#define error(x) freakout("WWW: " x)
#define POLLDELAY 1 /* seconds */
#define SCOREBOARD 10

pthread_t www_thread;
int refresh; /* saves me passing it around all the time */

static char cellbg_1[] = "\"#F0F0F0\"";
static char cellbg_2[] = "\"#E0E0E0\"";



inline void html_graph(content *cnt, char *title,
		int GRAPHWIDTH, int GRAPHHEIGHT,
		i64 graph_in[], i64 graph_out[], int pos, int GRAPH_SIZE)
{

#define _I_R 0xF0
#define _I_G 0xC0
#define _I_B 0x60

#define _O_R 0xA0
#define _O_G 0x50
#define _O_B 0x40

#define _IN "\"#F0C060\""
#define _OUT "\"#A05040\""

	int i, total_width = 0, total_strips = 0;
	double divisor;
	i64 max_transfer;
#ifndef HAVE_QUAD
	char *maximum;
#endif

	SET64(max_transfer, 0,0);

	/* find maximum value for this graph */
	for (i=0; i<GRAPH_SIZE; i++)
	{
		i64 tmp;
		
		SET64(tmp, HI(graph_in[i])+HI(graph_out[i]), LO(graph_in[i]));
		add_64bit(tmp, LO(graph_out[i]));
		max_transfer = max_64(max_transfer, tmp);
	}

#ifdef HAVE_QUAD
	c_appendf(cnt,
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">"
		"<tr><td bgcolor=\"#000000\">"
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">"
		"<tr><td align=\"center\">"
		"<font color=\"#FFFFFF\"><b>%s</b> "
		"(<b><font color="_IN">in</font></b> and "
		 "<b><font color="_OUT">out</font></b>)<br>"
		"Maximum: %'Lu bytes</td></tr>"
		"<tr><td bgcolor=\"#FFFFFF\">"
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\"><tr>",
		title, max_transfer);
#else
	maximum = add_commas( bignum_from_64bit(max_transfer), 1 );

	c_appendf(cnt,
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">"
		"<tr><td bgcolor=\"#000000\">"
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">"
		"<tr><td align=\"center\">"
		"<font color=\"#FFFFFF\"><b>%s</b> "
		"(<b><font color="_IN">in</font></b> and "
		 "<b><font color="_OUT">out</font></b>)<br>"
		"Maximum: %s bytes</td></tr>"
		"<tr><td bgcolor=\"#FFFFFF\">"
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\"><tr>",
		title, maximum);

	free(maximum);
#endif

	divisor = max(1.0, db_div_64_d(max_transfer, GRAPHHEIGHT));

	i = pos;
	do {
		int in, out, blank, taint = 0, width;

		i++;
		if (i == GRAPH_SIZE) i = 0;

		/* width calculation */
		width = (GRAPHWIDTH - total_width - 2*GRAPH_SIZE) /
			(GRAPH_SIZE - total_strips);
		total_strips++;
		total_width += width;

		/* get data */
		in = i_div_64_db(graph_in[i], divisor);
		out = i_div_64_db(graph_out[i], divisor);
		blank = GRAPHHEIGHT - in - out;

		/* write it out */
		c_appends(cnt, "<td valign=\"bottom\">");
		if (blank)
		{
			taint = 1;
			c_appendf(cnt,
			"<img src=\"i/x\" width=\"%d\" height=\"%d\">",
			width, blank);
		}
		if (in) c_appendf(cnt,
			"%s<img src=\"i/i\" width=\"%d\" height=\"%d\">",
				(blank)?"<br>":"", width, in);
		if (out) c_appendf(cnt,
			"%s<img src=\"i/o\" width=\"%d\" height=\"%d\">",
				(in||blank)?"<br>":"", width, out);

		c_appends(cnt, "</td>");

	} while (i != pos);

	c_appends(cnt,
		"</tr></table>"
		"</td></tr></table></td></tr></table>");
}



const char css[] =
	"<style type=\"text/css\">"
	"A{color:#404040;text-decoration:none;}"
	"A:hover{color:#000000;background:#E0E0E0; text-decoration:underline;}"
	"A.black{color:#FFFFFF;text-decoration:underline;}"
	"A.black:hover{color:#FFFFFF;background:#808080;"
	              "text-decoration:underline;}"
	"</style>";

const char linkbox[] =
	"<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">"
	"<tr><td bgcolor=\"#000000\"><font color=\"#FFFFFF\"><b>"
	"&nbsp;" PACKAGE " v" VERSION "&nbsp;"
	"</b></font></td>"
	"<td bgcolor=\"#000000\">"
	"<table border=\"0\" cellspacing=\"1\" cellpadding=\"2\">"
	"<tr><td bgcolor=\"#FFFFFF\">"
	"<a href=\"/\">main</a> | "
	"<a href=\"hosts-total.html\">hosts</a> | "
	"<a href=\"ports-total.html\">ports</a> | "
	"<a href=\"protocols.html\">protocols</a> | "
	"<a href=\"graphs.html\">graphs</a> | "
	"<a href=\"" URL "\">homepage</a>"
	"</td></tr></table>"
	"</td></tr></table>";

/* refresh idea due to Danny Brewer - thank you */
void refresh_header(content *cnt, char *title)
{
	if (refresh) c_appendf(cnt,
		"<html><head><title>darkstat : %s</title>"
		"<meta http-equiv=\"refresh\" content=\"%d\">"
		"%s</head><body>%s<br>",
		title, refresh, css, linkbox);

	else c_appendf(cnt,
		"<html><head><title>darkstat : %s</title>"
		"%s</head><body>%s<br>",
		title, css, linkbox);
}

void refresh_footer(content *cnt)
{
	if (refresh == 0) c_appends(cnt,
	"<form method=\"GET\">Want this page refreshed every "
	"<input type=\"text\" name=\"refresh\" value=\"5\" size=\"4\"> "
	"seconds? <input type=\"submit\" value=\"Yes\"></form>"
	"</body></html>");

	else c_appendf(cnt,
	"<form method=\"GET\">This page is being refreshed every "
	"<input type=\"text\" name=\"refresh\" value=\"%d\" size=\"4\"> "
	"seconds. <input type=\"submit\" value=\"Change\"> "
	"<input type=\"submit\" value=\"Stop\" name=\"cutitout\"></form>"
	"</body></html>",
	refresh);
}



content *www_frontpage(SOCKET incoming)
{
#ifndef HAVE_QUAD
	char *x_packets, *x_data;
#endif
	extern i64 num_packets, total_data;
	extern time_t t_start, t_already;
	content *cnt = c_new();

	time_t uptime = t_already + (time(NULL) - t_start);
	dword days, hours, minutes;
	
	refresh_header(cnt, "main");
	c_appends(cnt,
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">");

#ifdef HAVE_QUAD
	c_appendf(cnt,
	"<tr><td align=\"right\"><b>Packets captured:</b></td>"
	"<td>%'Lu</td></tr>"
	"<tr><td align=\"right\"><b>Data off the wire:</b></td>"
	"<td>%'Lu bytes</td></tr>", num_packets, total_data);
#else	
	/* packets */
	x_packets = add_commas( bignum_from_64bit(num_packets), 1 );
	c_appendf(cnt,
		"<tr><td align=\"right\"><b>Packets captured:</b></td>"
		"<td>%s</td></tr>", x_packets);
	free(x_packets);

	/* data */
	x_data = add_commas( bignum_from_64bit(total_data), 1 );
	c_appendf(cnt,
		"<tr><td align=\"right\"><b>Data off the wire:</b></td>"
		"<td>%s bytes</td></tr>", x_data);
	free(x_data);
#endif

	/* uptime */
#define S_MIN 60
#define S_HR (S_MIN*60)
#define S_DAY (S_HR*24)

	days = uptime / S_DAY;
	uptime -= days * S_DAY;

	hours = uptime / S_HR;
	uptime -= hours * S_HR;

	minutes = uptime / S_MIN;
	uptime -= minutes * S_MIN;

#undef S_DAY
#undef S_HR
#undef S_MIN

	c_appends(cnt, "<tr><td align=\"right\"><b>Counting:</b></td><td>");
	if (days)
		c_appendf(cnt, "%u days ", days);
	if (days || hours)
		c_appendf(cnt, "%u hours ", hours);
	if (days || hours || minutes)
		c_appendf(cnt, "%u minutes ", minutes);
	c_appendf(cnt, "%u seconds</td></tr>", uptime);

	/* since */
	c_appendf(cnt,
	"<tr><td align=\"right\"><b>Running since:</b></td><td>%s</td></tr>"
	"</table><br>",
	asctime(localtime(&t_start)) );

	pthread_mutex_lock(&graph_mutex);
	html_graph(cnt, "Last 60 seconds", 400, 200,
			graph_sec_in, graph_sec_out, pos_sec, GRAPH_SECS);
	pthread_mutex_unlock(&graph_mutex);

	c_appendf(cnt,
		"<br>DNS resolution is ::[ "
		"%s<a href=\"/dns-on.html\">on</a>%s | "
		"%s<a href=\"/dns-off.html\">off</a>%s | "
		"%s<a href=\"/dns-cycle.html\">cycling once</a>%s "
		"]::<br>",
(want_resolve==ON)?"<b>":"",(want_resolve==ON)?"</b>":"",
(want_resolve==OFF)?"<b>":"",(want_resolve==OFF)?"</b>":"",
(want_resolve==IN_PROGRESS)?"<b>":"",(want_resolve==IN_PROGRESS)?"</b>":""
		);

	refresh_footer(cnt);

	return cnt;
}



content *www_graphs(SOCKET incoming)
{
	content *cnt = c_new();

	refresh_header(cnt, "graphs");
	c_appends(cnt,
		"<table border=\"0\">"
		"<tr><td align=\"center\">");

	pthread_mutex_lock(&graph_mutex);
	html_graph(cnt, "Last 60 seconds", 300, 200,
			graph_sec_in, graph_sec_out, pos_sec, GRAPH_SECS);
	c_appends(cnt, "</td><td align=\"center\">");
	html_graph(cnt, "Last 60 minutes", 300, 200,
			graph_min_in, graph_min_out, pos_min, GRAPH_MINS);
	c_appends(cnt, "</td></tr><tr><td align=\"center\">");
	html_graph(cnt, "Last 24 hours", 300, 200,
			graph_hr_in, graph_hr_out, pos_hr, GRAPH_HRS);
	c_appends(cnt, "</td><td align=\"center\">");
	html_graph(cnt, "Last 30 days", 300, 200,
			graph_day_in, graph_day_out, pos_day, GRAPH_DAYS);
	pthread_mutex_unlock(&graph_mutex);

	c_appends(cnt, "</td></tr></table>");
	refresh_footer(cnt);

	return cnt;
}



content *www_hosts(SOCKET incoming, sort_type st, int limited)
{
	int i;
	content *cnt = c_new();
	host_record **list;
	int listend;
	char *cellbg = cellbg_2;

	refresh_header(cnt, "hosts");
	c_appendf(cnt,
		"Hosts (sorted by %s",
		(st==IN)?"incoming":(st==OUT)?"outgoing":
		(st==TOTAL)?"total":(st==MAIN)?"IP":"ERROR");

	pthread_mutex_lock(&db_mutex);
	if (limited)
		c_appendf(cnt, ", top %d",
			min(HOST_REPORT_LIMIT, host_db.used));

	list = host_db_sort(st);
	pthread_mutex_unlock(&db_mutex);

	c_appends(cnt, ")<br><br>"
	"<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\">"
	"<tr><td bgcolor=\"#000000\">"
	"<table border=\"0\" cellspacing=\"0\" cellpadding=\"2\"><tr>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"hosts-ip.html\">IP</a></b>"
	 " (<a class=\"black\" href=\"hosts-ip-full.html\">full</a>)"
	"</font></td>"
	"<td><font color=\"#FFFFFF\"><b>Hostname</b></font></td>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"hosts-in.html\">In</a></b>"
	 " (<a class=\"black\" href=\"hosts-in-full.html\">full</a>)"
	"</font></td>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"hosts-out.html\">Out</a></b>"
	 " (<a class=\"black\" href=\"hosts-out-full.html\">full</a>)"
	"</font></td>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"hosts-total.html\">Total</a></b>"
	 " (<a class=\"black\" href=\"hosts-total-full.html\">full</a>)"
	"</font></td>"
	"</tr>");

	/* you want the full thing or not? */
	if (limited)
		listend = host_db.used-HOST_REPORT_LIMIT;
	else
		listend = 0;
	if (listend < 0) listend = 0;

	for (i=host_db.used-1; i>=listend; i--)
	{
		host_record *h;
#ifndef HAVE_QUAD
		char *data_in, *data_out, *total;
		i64 t64;
#endif

		cellbg = (cellbg == cellbg_1) ? cellbg_2 : cellbg_1;

		/* grab the record */	
		h = list[i];

#ifdef HAVE_QUAD
		c_appendf(cnt,
			"<tr>"
			"<td bgcolor=%s>%d.%d.%d.%d</td>"
			"<td bgcolor=%s>%s</td>"
			"<td align=\"right\" bgcolor=%s>%'Lu</td>"
			"<td align=\"right\" bgcolor=%s>%'Lu</td>"
			"<td align=\"right\" bgcolor=%s>%'Lu</td></tr>",
			cellbg, (h->ip_addr >> 24) & 255,
				(h->ip_addr >> 16) & 255,
				(h->ip_addr >> 8) & 255,
				 h->ip_addr & 255,
			cellbg, h->hostname?h->hostname:"&nbsp;",
			cellbg, h->data_in,
			cellbg, h->data_out,
			cellbg, h->data_in + h->data_out);
#else
		data_in = add_commas( bignum_from_64bit(h->data_in), 1);
		data_out = add_commas( bignum_from_64bit(h->data_out), 1);

		t64 = h->data_in;
		add_64bit(t64, h->data_out.lo);
		t64.hi += h->data_out.hi;
		total = add_commas( bignum_from_64bit(t64), 1);

		c_appendf(cnt,
			"<tr>"
			"<td bgcolor=%s>%d.%d.%d.%d</td>"
			"<td bgcolor=%s>%s</td>"
			"<td align=\"right\" bgcolor=%s>%s</td>"
			"<td align=\"right\" bgcolor=%s>%s</td>"
			"<td align=\"right\" bgcolor=%s>%s</td></tr>",
			cellbg, (h->ip_addr >> 24) & 255,
				(h->ip_addr >> 16) & 255,
				(h->ip_addr >> 8) & 255,
				 h->ip_addr & 255,
			cellbg, h->hostname?h->hostname:"&nbsp;",
			cellbg, data_in,
			cellbg, data_out,
			cellbg, total);

		free(data_in);
		free(data_out);
		free(total);
#endif
	}
	free(list);

	c_appends(cnt,
		"</table>"
		"</td></tr></table>");

	refresh_footer(cnt);
	return cnt;
}



content *www_ports(SOCKET incoming, sort_type st, int limited)
{
	content *cnt = c_new();
	int i;
	port_record **list;
	int listend;
	char *cellbg = cellbg_2;

	refresh_header(cnt, "ports");
	c_appendf(cnt,
		"Ports (TCP, sorted by %s",
		(st==IN)?"incoming":(st==OUT)?"outgoing":
		(st==TOTAL)?"total":(st==MAIN)?"port number":"ERROR");

	pthread_mutex_lock(&db_mutex);

	if (limited)
		c_appendf(cnt, ", top %d",
			min(PORT_REPORT_LIMIT, port_db.used));
		
	c_appends(cnt, ")<br><br>"
	"<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\">"
	"<tr><td bgcolor=\"#000000\">"
	"<table border=\"0\" cellspacing=\"1\" cellpadding=\"2\">"
	"<tr>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"ports-num.html\">Port</a></b>"
	 " (<a class=\"black\" href=\"ports-num-full.html\">full</a>)"
	"</font></td>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"ports-in.html\">In</a></b>"
	 " (<a class=\"black\" href=\"ports-in-full.html\">full</a>)"
	"</font></td>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"ports-out.html\">Out</a></b>"
	 " (<a class=\"black\" href=\"ports-out-full.html\">full</a>)"
	"</font></td>"
	"<td><font color=\"#FFFFFF\">"
	 "<b><a class=\"black\" href=\"ports-total.html\">Total</a></b>"
	 " (<a class=\"black\" href=\"ports-total-full.html\">full</a>)"
	"</font></td>"
	"</tr>");

	list = port_db_sort(st);

	/* you want the full thing or not? */
	if (limited)
		listend = port_db.used-PORT_REPORT_LIMIT;
	else
		listend = 0;
	if (listend < 0) listend = 0;

	for (i=port_db.used-1; i>=listend; i--)
	{
		port_record *p;
#ifndef HAVE_QUAD
		char *data_in, *data_out, *total;
#endif

		cellbg = (cellbg == cellbg_1) ? cellbg_2 : cellbg_1;

		/* grab the record */	
		p = list[i];

#ifdef HAVE_QUAD
		c_appendf(cnt,
			"<tr><td bgcolor=%s>%d</td>"
			"<td align=\"right\" bgcolor=%s>%'Lu</td>"
			"<td align=\"right\" bgcolor=%s>%'Lu</td>"
			"<td align=\"right\" bgcolor=%s>%'Lu</td></tr>",
			cellbg,p->port, cellbg,p->data_in, cellbg,p->data_out,
			cellbg,p->data_in + p->data_out);
#else
		data_in = bignum_from_64bit(p->data_in);
		data_out = bignum_from_64bit(p->data_out);
		total = add_commas( add_bignum(data_in, data_out), 1);
		data_in = add_commas(data_in, 1);
		data_out = add_commas(data_out, 1);

		c_appendf(cnt,
			"<tr><td bgcolor=%s>%d</td>"
			"<td align=\"right\" bgcolor=%s>%s</td>"
			"<td align=\"right\" bgcolor=%s>%s</td>"
			"<td align=\"right\" bgcolor=%s>%s</td></tr>",
			cellbg,p->port, cellbg,data_in, cellbg,data_out,
			cellbg,total);

		free(data_in);
		free(data_out);
		free(total);
#endif
	}
	pthread_mutex_unlock(&db_mutex);
	free(list);

	c_appends(cnt,
		"</table>"
		"</td></tr></table>");

	refresh_footer(cnt);
	return cnt;
}



content *www_protocols(SOCKET incoming)
{
	content *cnt = c_new();
	char *cellbg = cellbg_2;
	int i;
	i64 zero;
	SET64(zero, 0, 0);

	refresh_header(cnt, "protocols");
	c_appends(cnt,
	"<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\">"
	"<tr><td bgcolor=\"#000000\">"
	"<table border=\"0\" cellspacing=\"1\" cellpadding=\"2\"><tr>"
	"<td colspan=\"2\" bgcolor=\"#000000\">"
		"<font color=\"#FFFFFF\"><b>Protocol</b></font></td>"
	"<td bgcolor=\"#000000\">"
		"<font color=\"#FFFFFF\"><b>In</b></font></td>"
	"<td bgcolor=\"#000000\">"
		"<font color=\"#FFFFFF\"><b>Out</b></font></td>"
	"<td bgcolor=\"#000000\">"
		"<font color=\"#FFFFFF\"><b>Total</b></font></td></tr>");

	for (i=0; i<256; i++)
	if (bigger_64(proto_in[i], zero) || bigger_64(proto_out[i], zero))
	{
		cellbg = (cellbg == cellbg_1) ? cellbg_2 : cellbg_1;

#ifdef HAVE_QUAD
	c_appendf(cnt,
		"<tr><td bgcolor=%s align=\"right\">%d</td>"
		"<td bgcolor=%s>%s</td>"
		"<td align=\"right\" bgcolor=%s>%'Lu</td>"
		"<td align=\"right\" bgcolor=%s>%'Lu</td>"
		"<td align=\"right\" bgcolor=%s>%'Lu</td></tr>\n",
		cellbg,i, cellbg,proto_name(i),
		cellbg,proto_in[i],
	        cellbg,proto_out[i],
		cellbg,proto_in[i] + proto_out[i]);
#else /* !HAVE_QUAD */
{
	char *data_in, *data_out, *total;
	data_in = bignum_from_64bit(proto_in[i]);
	data_out = bignum_from_64bit(proto_out[i]);
	total = add_commas( add_bignum(data_in, data_out), 1);
	data_in = add_commas(data_in, 1);
	data_out = add_commas(data_out, 1);
	c_appendf(cnt,
		"<tr><td bgcolor=%s align=\"right\">%d</td>"
		"<td bgcolor=%s>%s</td>"
		"<td align=\"right\" bgcolor=%s>%s</td>"
		"<td align=\"right\" bgcolor=%s>%s</td>"
		"<td align=\"right\" bgcolor=%s>%s</td></tr>\n",
		cellbg,i, cellbg,proto_name(i),
		cellbg,data_in, cellbg,data_out, cellbg,total);
	free(data_in);
	free(data_out);
	free(total);
}
#endif /* HAVE_QUAD */
	}

	c_appends(cnt,
		"</table>"
		"</td></tr></table>");

	refresh_footer(cnt);
	return cnt;
}



inline content *new_gif(void)
{
	content *cnt = c_new();

	assert(GIF_SIZE <= cnt->pool);
	memcpy(cnt->buf, gif, GIF_SIZE);
	cnt->used = GIF_SIZE;
	return cnt;
}



content *www_unimplemented()
{
	content *c = c_new();
	c_appends(c, "<html><body>"
		"<h1>Not Implemented</h1>"
		"Whatever the heck you just requested, I can't generate."
		"</body></html>");

	return c;
}



content *www_notfound()
{
	content *c = c_new();
	c_appends(c, "<html><body>"
		"<h1>Not Found</h1>"
		"Whatever you just requested, I don't have."
		"</body></html>");

	return c;
}



content *www_dnsack()
{
	content *c = c_new();
	c_appends(c,
		"<html><head>"
		 "<meta http-equiv=\"REFRESH\" content=\"0; URL=/\">"
		"</head>"
		"<body><h1>DNS control</h1>"
		"Done and done, now <a href=\"/\">back</a> you go."
		"</body></html>");
	return c;
}



SOCKET	sb_socket[SCOREBOARD];
int	sb_state[SCOREBOARD];
#define STATE_OPEN 0
#define STATE_REQUEST 1
#define STATE_REQCOMPLETE 2
#define STATE_REPLY 3

#define MAX_REQ_SIZE 8192
char	sb_request[SCOREBOARD][MAX_REQ_SIZE];
content *sb_reply[SCOREBOARD];
int	sb_sent[SCOREBOARD];

void kill_connection(int slot)
{
	if (sb_socket[slot] != -1)
	{
		shutdown(sb_socket[slot], SHUT_WR);
		close(sb_socket[slot]);
	}
	sb_socket[slot] = -1;
	sb_state[slot] = STATE_OPEN;
	sb_request[slot][0] = 0;
	if (sb_reply[slot]) c_delete(&(sb_reply[slot]));
	assert(sb_reply[slot] == NULL);
	sb_sent[slot] = 0;
}

int get_url(char *request, char *url)
{
	char d = request[strlen(url)];
	
	if (memcmp(request, url, strlen(url)) != 0) return 0;
	if (d == '\r' || d == '\n' || d == ' ' || d == '?') return 1;
	else return 0;
}

void handle_connection(int slot)
{
	content *hdr, *cnt;
	char *request = sb_request[slot], *tmp, *tmp2;
	int pos, recvd, is_gif = 0;
	SOCKET incoming = sb_socket[slot];

	assert(incoming != -1);

	/* FIXME: check for connections that time out */

	/* first, check if it's requesting */
	if (sb_state[slot] == STATE_REQUEST)
	{
		pos = strlen(request);

		recvd = recv(incoming, request+pos, MAX_REQ_SIZE-1-pos,
			MSG_NOSIGNAL);
		request[pos+recvd] = 0;

		if (recvd == MAX_REQ_SIZE-1-pos)
		{
			if (verbose)
				printf("WWW(%d): Request %d exceeded "
					"size of %d.\n",
					getpid(), slot, MAX_REQ_SIZE);

			kill_connection(slot);
			return;
		}
		if (recvd < 1)
		{
			if (verbose)
				printf("WWW(%d): Request %d broke off.\n",
					getpid(), slot);

			kill_connection(slot);
			return;
		}
	
		if (
			(strcmp(request+strlen(request)-2, "\n\n") == 0) ||
			(strcmp(request+strlen(request)-2, "\r\r") == 0) ||
			(strcmp(request+strlen(request)-4, "\r\n\r\n") == 0) ||
			(strcmp(request+strlen(request)-4, "\n\r\n\r") == 0) )
		{
			sb_state[slot] = STATE_REQCOMPLETE;
			shutdown(incoming, 0); /* no more recv */

#define CMP(x) (memcmp(request,x,strlen(x)) == 0)
#define GET(x) (get_url(request, "GET " x))

	hdr = c_new();

	/* parse ?refresh= in URL */
	refresh = 0;
	tmp = strstr(request, "Referer: ");
	if (!tmp) tmp = request + MAX_REQ_SIZE; /* cheater */
	tmp2 = strstr(request, "?refresh=");
	if (tmp2 && tmp2 < tmp) refresh = atoi(tmp2+9);
	tmp2 = strstr(request, "cutitout=Stop");
	if (tmp2 && tmp2 < tmp) refresh = 0;

	/* parse URL */
	if (!CMP("GET ")) {
		c_appends(hdr, "HTTP/1.1 501 Not Implemented\r\n");
		cnt = www_unimplemented();
	}
	else if GET("/")
			cnt = www_frontpage(incoming);
	else if GET("/hosts-total.html")
			cnt = www_hosts(incoming,TOTAL,1);
	else if GET("/hosts-total-full.html")
			cnt = www_hosts(incoming,TOTAL,0);
	else if GET("/hosts-ip.html")
			cnt = www_hosts(incoming,MAIN,1);
	else if GET("/hosts-ip-full.html")
			cnt = www_hosts(incoming,MAIN,0);
	else if GET("/hosts-in.html")
			cnt = www_hosts(incoming,IN,1);
	else if GET("/hosts-in-full.html")
			cnt = www_hosts(incoming,IN,0);
	else if GET("/hosts-out.html")
			cnt = www_hosts(incoming,OUT,1);
	else if GET("/hosts-out-full.html")
			cnt = www_hosts(incoming,OUT,0);
	else if GET("/ports-total.html")
			cnt = www_ports(incoming,TOTAL,1);
	else if GET("/ports-total-full.html")
			cnt = www_ports(incoming,TOTAL,0);
	else if GET("/ports-num.html")
			cnt = www_ports(incoming,MAIN,1);
	else if GET("/ports-num-full.html")
			cnt = www_ports(incoming,MAIN,0);
	else if GET("/ports-in.html")
			cnt = www_ports(incoming,IN,1);
	else if GET("/ports-in-full.html")
			cnt = www_ports(incoming,IN,0);
	else if GET("/ports-out.html")
			cnt = www_ports(incoming,OUT,1);
	else if GET("/ports-out-full.html")
			cnt = www_ports(incoming,OUT,0);
	else if GET("/protocols.html")
			cnt = www_protocols(incoming);
	else if GET("/graphs.html")
			cnt = www_graphs(incoming);
	else if CMP("GET /i/")
	{
		/* "validator" */
		if (strstr(request,"If-Modified-Since:"))
		{
			sb_reply[slot] = hdr;
			c_appends(sb_reply[slot],
				"HTTP/1.1 304 Not Modified\r\n");
			sb_state[slot] = STATE_REPLY;
			return;
		}

		cnt = new_gif();
		is_gif = 1;

		if GET("/i/i")
		{
			cnt->buf[GIF_RED] = _I_R;
			cnt->buf[GIF_GREEN] = _I_G;
			cnt->buf[GIF_BLUE] = _I_B;
		}
		else if GET("/i/o")
		{
			cnt->buf[GIF_RED] = _O_R;
			cnt->buf[GIF_GREEN] = _O_G;
			cnt->buf[GIF_BLUE] = _O_B;
		}
		else
			cnt->buf[GIF_PIXEL] = GIF_TRANSPARENT;
	}
	else if CMP("GET /dns-")
	{
		c_appends(hdr,
			"HTTP/1.1 302 Found\r\n"
			"Location: /\r\n");

		if GET("/dns-on.html") want_resolve = ON;
		else if GET("/dns-off.html") want_resolve = OFF;
		else if GET("/dns-cycle.html") want_resolve = IN_PROGRESS;

		cnt = www_dnsack();
	} else {
		c_appends(hdr, "HTTP/1.1 404 Not Found\r\n");
		cnt = www_notfound();
	}
#undef GET
#undef CMP
			if (!hdr->used) c_appends(hdr, "HTTP/1.1 200 OK\r\n");
			if (is_gif)
			{
				char datebuf[200];
				time_t clock = time(NULL);
				struct tm *tm = gmtime(&clock);
				strftime(datebuf, 200,
					"%a, %e %b %Y %H:%M:%S GMT", tm);
				c_appendf(hdr,
					"Date: %s\r\n"
					"Last-Modified: %s\r\n",
					datebuf, datebuf);

				c_appends(hdr,
				 "Cache-Control: max-age=86400, public\r\n"
				 "Content-Type: image/gif\r\n");
			}
			else
				c_appends(hdr,
				 "Cache-Control: no-cache, must-revalidate, "
				                "max-age=0\r\n"
				 "Content-Type: text/html\r\n");

			c_appendf(hdr,	"Content-Length: %d\r\n"
					"Connection: close\r\n"
					"\r\n",
					cnt->used);

			/* build final content buffer */
			sb_reply[slot] = hdr;
			sb_sent[slot] = 0;
			c_appendsl(sb_reply[slot], cnt->buf, cnt->used);
			c_delete(&cnt);

			/* we can now start sending it */
			sb_state[slot] = STATE_REPLY;
		}

		/* end of requesting code */
		return;
	}

	/* it's not requesting, it could be replying */
	if (sb_state[slot] == STATE_REPLY)
	{
		int sent = send(incoming,
			sb_reply[slot]->buf + sb_sent[slot],
			sb_reply[slot]->used - sb_sent[slot],
			MSG_NOSIGNAL);

		if (sent < 1)
		{
			if (verbose)
				printf("WWW(%d): Slot %d died during "
					"reply sending.\n",
					getpid(), slot);

			kill_connection(slot);
			return;
		}

		/* is it finished? */
		if (sent == sb_reply[slot]->used - sb_sent[slot])
			kill_connection(slot);
		else
		{
			if (verbose)
				printf("sent %d+%d, used %d\n",
					sb_sent[slot], sent,
					sb_reply[slot]->used);
			sb_sent[slot] += sent;
		}
	}
}



RETSIGTYPE broken_pipe(int signum)
{
	if (verbose)
		fprintf(stderr, "WWW: Suppressed SIGPIPE.\n");
}



SOCKET sockin;

void poll(void)
{
	int i, max_fd, sb_open = -1;
	struct timeval timeout;
	fd_set r_fd, w_fd;

	max_fd = sockin;
	FD_ZERO(&r_fd);
	FD_ZERO(&w_fd);

	for (i=0; i<SCOREBOARD; i++)
	{
		if (sb_socket[i] != -1)
			max_fd = max(max_fd, sb_socket[i]);
		else
			sb_open = i;
		
		/* poll connected sockets in REQUEST for reading */
		if (sb_state[i] == STATE_REQUEST)
			FD_SET(sb_socket[i], &r_fd);

		/* REPLY for writing */
		if (sb_state[i] == STATE_REPLY)
			FD_SET(sb_socket[i], &w_fd);

	}

	/* if scoreboard has an opening, poll sockin */
	if (sb_open != -1)
		FD_SET(sockin, &r_fd);

	timeout.tv_sec = POLLDELAY;
	timeout.tv_usec = 0;
	if (select(max_fd+1, &r_fd, &w_fd, NULL, &timeout) == -1)
		return; /* nothing worth doing */

	if (FD_ISSET(sockin, &r_fd))
	{
		struct sockaddr_in addrin;
		int sin_size;
		SOCKET incoming;

		assert(sb_open != -1);

		sin_size = sizeof(struct sockaddr);
		incoming = accept(sockin, (struct sockaddr *)&addrin,
			&sin_size);

		/* changed your mind? <-- happens often when the
		 * web interface is being hammered with requests
		 */
		if (incoming < 0)
		{
			if (verbose)
			printf("WWW: Lost a connection in accept()\n");
		}
		else
		{
			/* details about incoming connection */
			if (verbose)
				printf("WWW: Got a connection from %s:%d\n",
					inet_ntoa(addrin.sin_addr),
					ntohs(addrin.sin_port));

			sb_socket[sb_open] = incoming;
			sb_state[sb_open] = STATE_REQUEST;
		}
	}

	for (i=0; i<SCOREBOARD; i++)
	if (sb_socket[i] != -1)
		if (FD_ISSET(sb_socket[i], &r_fd) ||
			FD_ISSET(sb_socket[i], &w_fd))
				handle_connection(i);
}



void www_main(void *ptr)
{
	struct sockaddr_in addrin;
	int sockopt = 1, i;

	/* reset scoreboard */
	for (i=0; i<SCOREBOARD; i++)
	{
		sb_socket[i] = -1;
		sb_state[i] = STATE_OPEN;
		sb_request[i][0] = 0;
		sb_reply[i] = NULL;
		sb_sent[i] = 0;
	}

	/* create the incoming socket */
	sockin = socket(AF_INET, SOCK_STREAM, 0);
	if (sockin < 0) error("Problem creating socket");

	/* reuse address */
	if (setsockopt(sockin, SOL_SOCKET, SO_REUSEADDR,
		(char*)&sockopt, sizeof(sockopt)) < 0)
			error("WWW: Can't REUSEADDR");

	/* fill out a sockaddr struct */
	addrin.sin_family = AF_INET;
	addrin.sin_port = htons(webport);
	addrin.sin_addr.s_addr = INADDR_ANY;
	bzero(&(addrin.sin_zero), 8);

	/* bind sockin to the incoming port */
	if (bind(sockin, (struct sockaddr *)&addrin,
		sizeof(struct sockaddr)) < 0)
	{
		fprintf(stderr,"The web port is %d.\n", webport);
		
		if (webport < 1024)
			error("Problem binding socket to port < 1024.  "
				"Are you not root?");
		else
			error("Problem binding socket.  "
				"Is the web port taken?");
	}
	
	/* listen on the socket */
	if (listen(sockin, SCOREBOARD) < 0)
		error("Problem listen()ing to socket");

	printf("WWW: Thread is awake and awaiting connections.\n");

	if (!MSG_NOSIGNAL)
		/* sometimes send() and recv() throw it */
		signal(SIGPIPE, broken_pipe);

	while (!shutting_down) poll();

	close(sockin);
	for (i=0; i<SCOREBOARD; i++)
		kill_connection(i);

	printf("WWW: Thread down.\n");
	pthread_exit(0);
}

